@@ -1,4 +1,3 @@
-// -*- c++ -*-
 /*
  * Copyright 2003, The libsigc++ Development Team
  *
@@ -15,13 +14,25 @@
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
- *
  */
 #include <sigc++/signal_base.h>
+#include <memory> // std::unique_ptr
 
 namespace sigc {
 namespace internal {
 
+// Data sent from signal_impl::insert() to slot_rep::set_parent() when a slot is
+// connected, and then sent from slot_rep::disconnect() to signal_impl::notify()
+// when the slot is disconnected. Bug 167714.
+struct self_and_iter
+{
+  signal_impl* self_;
+  signal_impl::iterator_type iter_;
+
+  self_and_iter(signal_impl* self, signal_impl::iterator_type iter)
+    : self_(self), iter_(iter) {}
+};
+
 signal_impl::signal_impl()
 : ref_count_(0), exec_count_(0), deferred_(0)
 {}
@@ -41,6 +52,18 @@ void signal_impl::operator delete(void* p)
 
 void signal_impl::clear()
 {
+  // Don't let signal_impl::notify() erase the slots. It would invalidate the
+  // iterator in the following loop.
+  const bool saved_deferred = deferred_;
+  signal_exec exec(this);
+
+  // Disconnect all connected slots before they are deleted.
+  // signal_impl::notify() will be called and delete the self_and_iter structs.
+  for (auto& slot : slots_)
+    slot.disconnect();
+
+  deferred_ = saved_deferred;
+
   slots_.clear();
 }
 
@@ -51,9 +74,9 @@ signal_impl::size_type signal_impl::size() const
 
 bool signal_impl::blocked() const
 {
-  for (const_iterator_type iter = slots_.begin(); iter != slots_.end(); ++iter)
+  for (const auto& slot : const_cast<const std::list<slot_base>&>(slots_))
   {
-    if (!iter->blocked())
+    if (!slot.blocked())
       return false;
   }
   return true;
@@ -61,9 +84,9 @@ bool signal_impl::blocked() const
 
 void signal_impl::block(bool should_block)
 {
-  for (iterator_type iter = slots_.begin(); iter != slots_.end(); ++iter)
+  for (auto& slot : slots_)
   {
-    iter->block(should_block);
+    slot.block(should_block);
   }
 }
 
@@ -74,20 +97,37 @@ signal_impl::iterator_type signal_impl::connect(const slot_base& slot_)
 
 signal_impl::iterator_type signal_impl::erase(iterator_type i)
 {
+  // Don't let signal_impl::notify() erase the slot. It would be more
+  // difficult to get the correct return value from signal_impl::erase().
+  const bool saved_deferred = deferred_;
+  signal_exec exec(this);
+
+  // Disconnect the slot before it is deleted.
+  // signal_impl::notify() will be called and delete the self_and_iter struct.
+  i->disconnect();
+
+  deferred_ = saved_deferred;
+
   return slots_.erase(i);
 }
     
 signal_impl::iterator_type signal_impl::insert(signal_impl::iterator_type i, const slot_base& slot_)
 {
-  iterator_type temp = slots_.insert(i, slot_);
-  temp->set_parent(this, &notify);
+  auto temp = slots_.insert(i, slot_);
+  auto si = new self_and_iter(this, temp);
+  temp->set_parent(si, &notify);
   return temp;
 }
 
 void signal_impl::sweep()
 {
+  // The deletion of a slot may cause the deletion of a signal_base,
+  // a decrementation of ref_count_, and the deletion of this.
+  // In that case, the deletion of this is deferred to ~signal_exec().
+  signal_exec exec(this);
+
   deferred_ = false;
-  iterator_type i = slots_.begin();
+  auto i = slots_.begin();
   while (i != slots_.end())
     if ((*i).empty())
       i = slots_.erase(i);
@@ -95,20 +135,29 @@ void signal_impl::sweep()
       ++i;
 }
 
+//static
 void* signal_impl::notify(void* d)
 {
-  signal_impl* self = reinterpret_cast<signal_impl*>(d);
-  if (self->exec_count_ == 0)
-    self->sweep();
-  else                       // This is occuring during signal emission.
-    self->deferred_ = true;  // => sweep() will be called from ~signal_exec().
-  return 0;                  // This is safer because we don't have to care about our iterators in emit().
+  std::unique_ptr<self_and_iter> si(static_cast<self_and_iter*>(d));
+
+  if (si->self_->exec_count_ == 0)
+  {
+    // The deletion of a slot may cause the deletion of a signal_base,
+    // a decrementation of si->self_->ref_count_, and the deletion of si->self_.
+    // In that case, the deletion of si->self_ is deferred to ~signal_exec().
+    signal_exec exec(si->self_);
+    si->self_->slots_.erase(si->iter_);
+  }
+  else                           // This is occuring during signal emission or slot erasure.
+    si->self_->deferred_ = true; // => sweep() will be called from ~signal_exec() after signal emission.
+  return 0;                      // This is safer because we don't have to care about our
+                                 // iterators in emit(), clear(), and erase().
 }
 
 } /* namespace internal */
 
 signal_base::signal_base()
-: impl_(0)
+: impl_(nullptr)
 {}
 
 signal_base::signal_base(const signal_base& src)
@@ -121,7 +170,14 @@ signal_base::signal_base(const signal_base& src)
 signal_base::~signal_base()
 {
   if (impl_)
+  {
+    // Disconnect all slots before impl_ is deleted.
+    // TODO: Move the signal_impl::clear() call to ~signal_impl() when ABI can be broken.
+    if (impl_->ref_count_ == 1)
+      impl_->clear();
+
     impl_->unreference();
+  }
 }
 
 void signal_base::clear()
@@ -169,7 +225,17 @@ signal_base::iterator_type signal_base::erase(iterator_type i)
 
 signal_base& signal_base::operator = (const signal_base& src)
 {
-  if (impl_) impl_->unreference();
+  if (src.impl_ == impl_) return *this;
+
+  if (impl_)
+  {
+    // Disconnect all slots before impl_ is deleted.
+    // TODO: Move the signal_impl::clear() call to ~signal_impl() when ABI can be broken.
+    if (impl_->ref_count_ == 1)
+      impl_->clear();
+
+    impl_->unreference();
+  }
   impl_ = src.impl();
   impl_->reference();
   return *this;
